#  Copyright (C) 2024
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from typing import TypeVar, Any, Callable

from .._type.is_iterable import is_iterable, ClassInfo, DEFAULT_BASE_TYPES

T = TypeVar('T', bound=Any)
P = TypeVar('P', bound=Any)


def to_tuple(
        x: T, base_type: ClassInfo | None = DEFAULT_BASE_TYPES, none: P | Callable[[], P] = tuple
) -> tuple[T, ...] | P:
    """If *obj* is iterable, convert to tuple::

        >>> to_tuple([1, 2, 3])
        (1, 2, 3)

    If *obj* is not iterable, wrap it in tuple::

        >>> to_tuple(1)
        (1,)

    If *obj* is ``None``, return *none*, by default empy tuple:

        >>> to_tuple(None)
        ()

    *none* could be set to literal or callable:

        >>> assert to_tuple(None, none=None) is None
        >>> to_tuple(None, none=lambda: (None, ))
        (None,)

    By default, binary and text strings, dicts, generic aliases are not considered iterable:

        >>> to_tuple('foo')
        ('foo',)

    If *base_type* is set, objects for which ``isinstance(obj, base_type)``
    returns ``True`` won't be considered iterable.

        >>> to_tuple([1, 2])  # list considered iterable by default
        (1, 2)
        >>> to_tuple([1, 2], base_type=(list,))  # Treat list as a unit
        ([1, 2],)

    Set *base_type* to ``None`` to avoid any special handling and treat objects
    Python considers iterable as iterable:

        >>> to_tuple('foo', base_type=None)
        ('f', 'o', 'o')

    Based on always_iterable from more_itertools==11.0.1
    """
    if x is None:
        return none() if callable(none) else none

    if isinstance(x, tuple):
        return x

    if is_iterable(x, base_type=base_type):
        return tuple(x)
    return (x, )



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import types
import typing

note: typing.TypeAlias = typing.Annotated
TypingGenericAlias: typing.TypeAlias = type(typing.List[str])
AnnotatedAlias = type(typing.Annotated[int, ''])

T = typing.TypeVar("T")
MaybeAnnotated: typing.TypeAlias = T | typing.Annotated[T, typing.Any]
WideType: typing.TypeAlias = type | types.UnionType | types.GenericAlias | TypingGenericAlias  # OR typing.Annotated
AnyType: typing.TypeAlias = WideType | AnnotatedAlias



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from .str_to_bool import str_to_bool
from .to_bool import to_bool
from .to_datetime import to_datetime
from .to_list import to_list
from .to_tuple import to_tuple
from .try_cast import try_cast
from .unflatten_dict import unflatten_dict
from .url_to_snake_case import url_to_snake_case



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from typing import Any, TypeVar, Callable, overload

T = TypeVar('T')
F = TypeVar('F')


@overload
def try_cast(value: Any, target_type: Callable[[Any], T], fallback_value: None = None,
             fallback_on: type[Exception] = Exception) -> T | None: pass


@overload
def try_cast(value: Any, target_type: Callable[[Any], T], fallback_value: F,
             fallback_on: type[Exception] = Exception) -> T | F: pass


def try_cast(value: Any, target_type: Callable[[Any], T], fallback_value: F | None = None,
             fallback_on: type[Exception] = Exception) -> T | F | None:
    try:
        return target_type(value)
    except fallback_on:
        return fallback_value



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
#
import dataclasses
import types
import typing
from types import UnionType, NoneType
from typing import Union


def get_name_from_type(type_: type | types.UnionType | dataclasses.InitVar | None | NoneType) -> str:
    """
    Returns compact and unambiguous name of any basic or generic type
    >>> get_name_from_type(str)
    'str'
    >>> get_name_from_type(int)
    'int'
    >>> get_name_from_type(None)
    'None'
    >>> get_name_from_type(bool)
    'bool'
    >>> get_name_from_type(float)
    'float'
    >>> get_name_from_type(list)
    'list'
    >>> get_name_from_type(typing.List)
    'list'
    >>> get_name_from_type(list[int])
    'list[int]'
    >>> get_name_from_type(typing.List[int])
    'list[int]'
    >>> get_name_from_type(tuple)
    'tuple'
    >>> get_name_from_type(typing.Tuple)
    'tuple'
    >>> get_name_from_type(tuple[int])
    'tuple[int]'
    >>> get_name_from_type(typing.Tuple[int])
    'tuple[int]'
    >>> get_name_from_type(tuple[int, str])
    'tuple[int, str]'
    >>> get_name_from_type(typing.Tuple[int, str])
    'tuple[int, str]'
    >>> get_name_from_type(tuple[int, ...])
    'tuple[int, ...]'
    >>> get_name_from_type(typing.Tuple[int, ...])
    'tuple[int, ...]'
    >>> get_name_from_type(dict)
    'dict'
    >>> get_name_from_type(typing.Dict)
    'dict'
    >>> get_name_from_type(dict[str, int])
    'dict[str, int]'
    >>> get_name_from_type(typing.Dict[str, int])
    'dict[str, int]'
    >>> # noinspection PyTypeChecker
    >>> get_name_from_type(typing.Union)
    'Union'
    >>> get_name_from_type(typing.Union[str, int])
    'str | int'
    >>> get_name_from_type(str | int)
    'str | int'
    >>> get_name_from_type(typing.Optional[str])
    'str | None'
    >>> get_name_from_type(typing.Union[str, None])
    'str | None'
    >>> get_name_from_type(str | None)
    'str | None'
    >>> get_name_from_type(dataclasses.InitVar[bool])
    'bool'
    >>> get_name_from_type(dataclasses.InitVar[float])
    'float'
    >>> get_name_from_type(typing.Dict[typing.Tuple[typing.Union[int, str], ...], typing.Optional[list[str]]])
    'dict[tuple[int | str, ...], list[str] | None]'
    """

    type_origin = typing.get_origin(type_)
    type_args = typing.get_args(type_)
    if type_origin is not None and type_args is not None:
        if type_origin in (Union, UnionType):
            return ' | '.join(list(get_name_from_type(sub_type) for sub_type in type_args))
        result = get_name_from_type(type_origin)
        if type_args:
            result += '[' + (', '.join(list(get_name_from_type(sub_type) for sub_type in type_args))) + ']'
        return result
    elif isinstance(type_, dataclasses.InitVar):
        return get_name_from_type(type_.type)
    elif type_ in (None, NoneType):
        return "None"
    elif type_ is Ellipsis:
        return "..."
    return type_.__name__



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from typing import TypeVar, Any

from no_value import NoValue

from ._exceptions import UnsupportedType
from .get_converter import get_converter
from .._type import MaybeAnnotated, try_extract_type_notes

T = TypeVar('T')


def convert_to_type(type_: MaybeAnnotated[type[T]], value: Any | NoValue) -> T:
    # noinspection PyTypeChecker
    type_: type[T] = try_extract_type_notes(type_)[0]

    try:
        converter = get_converter(type_)
    except TypeError as e:
        raise UnsupportedType(repr(e)) from e

    value = converter(value)
    return value



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import dataclasses
import functools
from typing import Any

from frozendict import frozendict


@functools.cache
def get_dataclass_field_name_to_field(
        dataclass_: Any,
        with_init_vars: bool = True, with_class_vars: bool = False
) -> frozendict[str, dataclasses.Field]:
    try:
        # noinspection PyUnresolvedReferences,PyProtectedMember
        fields = getattr(dataclass_, dataclasses._FIELDS)
    except AttributeError:
        raise TypeError("expected dataclass type")
    # noinspection PyUnresolvedReferences,PyProtectedMember
    allowed_field_types = {dataclasses._FIELD}
    # noinspection PyUnresolvedReferences,PyProtectedMember
    allowed_field_types |= {dataclasses._FIELD_INITVAR} if with_init_vars else set()
    # noinspection PyUnresolvedReferences,PyProtectedMember
    allowed_field_types |= {dataclasses._FIELD_CLASSVAR} if with_class_vars else set()
    # noinspection PyProtectedMember
    result = {name: field for name, field in fields.items() if field._field_type in allowed_field_types}
    return frozendict(result)



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
#
import dataclasses
from collections.abc import Mapping
from typing import Any, TypeVar

from frozendict import frozendict
from no_value import NoValue
from typing_extensions import deprecated

from ._exceptions import FieldErrors
from .get_dataclass_field_name_to_field import get_dataclass_field_name_to_field
from .get_dataclass_field_name_to_type import get_dataclass_field_name_to_type
from .get_field_value import get_field_value
from .._type import DataclassProtocol, MaybeAnnotated, WideType, try_extract_type_notes

DP = TypeVar('DP', bound=DataclassProtocol)


def to_dataclass(
        value: Mapping[str, Any] | DP | object,
        dataclass_type: MaybeAnnotated[type[DP]],
        stop_on_first_error: bool = False
) -> DP:
    """
    Return a dataclass instance based on first argument:
        - dict will be used to initialize new dataclass instance
        - dataclass instance of same or child classes will be passed unchanged
        - any other object will be used to extract data by getattr

    Uses get_field_value to prepare data for dataclass initializer
    """
    dataclass_type: type[DP] = try_extract_type_notes(dataclass_type)[0]
    if not dataclasses.is_dataclass(dataclass_type):
        raise TypeError("expected dataclass type")

    if isinstance(value, dataclass_type):
        return value

    dataclass_kwargs: dict[str, Any] = {}
    field_name_to_field: frozendict[str, dataclasses.Field] = get_dataclass_field_name_to_field(dataclass_type)
    field_name_to_type: frozendict[str, WideType] = get_dataclass_field_name_to_type(dataclass_type)
    get_value = value.get if hasattr(value, "get") else lambda key, default: getattr(value, key, default)

    field_name_to_error: dict[str, Exception] = {}
    for field_name in field_name_to_field:
        field = field_name_to_field[field_name]
        field_type = field_name_to_type[field_name]
        raw_value = get_value(field_name, NoValue)
        try:
            result_value = get_field_value(
                type_=field_type, init=field.init, default=field.default,
                default_factory=field.default_factory, value=raw_value)
        except FieldErrors as e:
            field_name_to_error |= {f'{field_name}.{field}': error for field, error in e.field_to_error.items()}
            if stop_on_first_error:
                break
            continue
        except Exception as e:
            field_name_to_error[field_name] = e
            if stop_on_first_error:
                break
            continue

        if result_value is not NoValue:
            dataclass_kwargs[field_name] = result_value

    if field_name_to_error:
        raise FieldErrors(field_name_to_error)

    return dataclass_type(**dataclass_kwargs)


@deprecated("Use to_dataclass instead")
def dict_to_dataclass(
        dict_value: Mapping[str, Any] | DP | object, dataclass: MaybeAnnotated[type[DP]]
) -> DP:
    return to_dataclass(dict_value, dataclass)



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import datetime
from typing import Union


def to_datetime(value: Union[str, int, float]) -> datetime.datetime:
    if isinstance(value, str):
        return datetime.datetime.fromisoformat(value)
    return datetime.datetime.fromtimestamp(value, datetime.timezone.utc)



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from ._cast import *
from ._dataclass import *
from ._type import *



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from ._alias import TypingGenericAlias, AnnotatedAlias, MaybeAnnotated, WideType, AnyType, note
from .dataclass_protocol import DataclassProtocol
from .eval_forward_refs import eval_forward_refs_in_local_dataclasses, eval_type
from .eval_forward_refs import get_evaled_dataclass_fields, eval_dataclass_fields
from .get_container_type import get_container_type
from .get_name_from_type import get_name_from_type
from .get_non_generic_args import get_non_generic_args
from .is_collection import is_collection
from .is_instance import is_instance
from .is_iterable import is_iterable, ClassInfo, DEFAULT_BASE_TYPES
from .is_mapping import is_mapping
from .is_tuple import is_tuple
from .try_extract_type_notes import try_extract_type_notes



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>


def str_to_bool(bool_str: str) -> bool:
    true_values = {"true", "1", "on", "+", "yes"}
    false_values = {"false", "0", "off", "-", "no"}
    lowed_str = bool_str.lower()
    if lowed_str in true_values:
        return True
    if lowed_str in false_values:
        return False
    raise ValueError(f'invalid literal for boolean: "{lowed_str}"({type(lowed_str)})')



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import dataclasses
import datetime
import functools
import itertools
import json
import types
import typing
from typing import Callable, Any, Union

from .._cast import to_bool, to_datetime
from .._type import MaybeAnnotated, WideType, is_instance, try_extract_type_notes, get_container_type

T = typing.TypeVar('T')


def pass_value(value: Any) -> Any:
    return value


@functools.cache
def get_converter(type_: MaybeAnnotated[type[T]]) -> Callable[[Any], T]:
    type_: WideType[T] = try_extract_type_notes(type_)[0]
    type_origin = typing.get_origin(type_)
    type_args = typing.get_args(type_)
    if not type_origin and isinstance(type_, type):
        if dataclasses.is_dataclass(type_):
            from .to_dataclass import to_dataclass
            return functools.partial(to_dataclass, dataclass_type=type_)
        converter = type_
        if type_ == bool:
            converter = to_bool
        if type_ == typing.Any:
            converter = pass_value
        if type_ == datetime.time:
            converter = datetime.time.fromisoformat
        if type_ == datetime.date:
            converter = datetime.date.fromisoformat
        if type_ == datetime.datetime:
            converter = to_datetime
    elif type_origin is Union or isinstance(type_, types.UnionType):
        def union_converter(value: str):
            if is_instance(value, type_args):
                return value
            for possible_type in type_args:
                try:
                    return get_converter(possible_type)(value)
                except (TypeError, ValueError, AttributeError):
                    pass
            raise ValueError(f"Expected one of: {type_args}, got {type(value)}: {value}")

        return union_converter
    elif type_origin is not None:
        container_type = get_container_type(type_origin)
        if container_type == dict:
            key_converter = get_converter(type_args[0])
            value_converter = get_converter(type_args[1])

            def converter(base_value: Any):
                raw_values = json.loads(base_value) if isinstance(base_value, str) else base_value
                if not isinstance(raw_values, dict):
                    raise ValueError(f'Expected dict, got {type(raw_values).__name__}: {raw_values}')
                parsed_values = [(key_converter(key), value_converter(value)) for key, value in raw_values.items()]
                result = container_type(parsed_values)
                return result
        elif container_type == tuple:
            has_ellipsis = False
            element_converters = []
            for type_ in type_args:
                if type_ is ...:
                    has_ellipsis = True
                else:
                    if has_ellipsis:
                        raise TypeError("Ellipsis (three dots (...)) should be the last"
                                        " in tuple to make it variable-length")
                    element_converters.append(get_converter(type_))
            if has_ellipsis:
                element_converters = itertools.cycle(element_converters)

            def converter(base_value: Any):
                raw_values = json.loads(base_value) if isinstance(base_value, str) else base_value
                if not isinstance(raw_values, (list, tuple, set, frozenset)):
                    raise ValueError(f'Expected iterable, got {type(raw_values).__name__}: {raw_values}')
                if not has_ellipsis and len(raw_values) != len(element_converters):
                    raise ValueError(f'Expected exactly {len(element_converters)} elements, got {len(raw_values)}')
                result = tuple(converter_(element) for element, converter_ in zip(raw_values, element_converters))
                return result
        else:
            element_converter = get_converter(type_args[0])

            def converter(base_value: Any):
                raw_values = json.loads(base_value) if isinstance(base_value, str) else base_value
                if not isinstance(raw_values, (list, tuple, set, frozenset)):
                    raise ValueError(f'Expected iterable, got {type(raw_values).__name__}: {raw_values}')
                parsed_values = [element_converter(element) for element in raw_values]
                result = container_type(parsed_values)
                return result
    elif isinstance(type_, dataclasses.InitVar):
        return get_converter(type_.type)
    else:
        raise TypeError(f'Attempt to cast to unexpected type: {type_} ({type(type_).__name__})')
    return converter



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import collections.abc
import types
from typing import Any, Sequence, Union


def is_collection(value: Any, collection_args: Sequence[type | types.GenericAlias]) -> bool:
    if not isinstance(value, collections.abc.Collection):
        return False

    if not collection_args:
        return True

    if len(collection_args) != 1:
        raise TypeError(f"Collection requires exactly 1 type arg, got: {collection_args}")

    collection_type_list = [collection_args[0]]

    from .is_instance import is_instance
    for val in value:
        if not is_instance(val, collection_type_list):
            return False
    return True



#  Copyright (C) 2024
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import dataclasses
from typing import runtime_checkable, Protocol, ClassVar, Any


# PyCharm compatible, but incompatible with MyPy AND DOES NOT WORK in 3.14+
@runtime_checkable
@dataclasses.dataclass
class DataclassProtocol(Protocol):
    pass


# MyPy compatible, but incompatible with PyCharm
@runtime_checkable
class DataclassProtocol(Protocol):
    __dataclass_fields__: ClassVar[dict[str, Any]]



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import dataclasses
from typing import Any, Callable, TypeVar

from no_value import NoValue
from typing_extensions import deprecated

from ._exceptions import MissingField
from .convert_to_type import convert_to_type
from .._type import MaybeAnnotated, WideType, try_extract_type_notes

T = TypeVar('T')


def get_field_value(
        type_: MaybeAnnotated[type[T]],
        init: bool,
        default: Any,
        default_factory: Callable,
        value: Any,
) -> T:
    """
    Implement dataclass field logic init/default/default_factory.
    WARNING! with value == NoValue and is_dataclass(type_): will try to initialize type_ instance from empty dict
    (it's useful in cascade configs parsing)
    Uses convert_to_type to convert value.
    """
    if not init:
        return NoValue

    type_: WideType = try_extract_type_notes(type_)[0]
    if value != NoValue:
        return convert_to_type(type_, value)
    if default != dataclasses.MISSING:
        return NoValue
    if default_factory != dataclasses.MISSING:
        return NoValue

    # attempt to init dataclass
    if dataclasses.is_dataclass(type_):
        return convert_to_type(type_, {})

    raise MissingField()


@deprecated('Use get_field_value instead')
def convert(
        type_: MaybeAnnotated[type[T]],
        init: bool,
        default: Any,
        default_factory: Callable,
        value: Any,
) -> T:
    return get_field_value(type_=type_, init=init, default=default, default_factory=default_factory, value=value)



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import types
import typing

ClassInfo: typing.TypeAlias = type | types.UnionType | tuple['ClassInfo', ...]
DEFAULT_BASE_TYPES: tuple[type, ...] = (str, bytes, dict, type(typing.List[int]), type(list[int]))


def is_iterable(obj: typing.Any, base_type: ClassInfo | None = DEFAULT_BASE_TYPES) -> bool | None:
    """If *obj* is iterable, return True::

        >>> is_iterable([1, 2, 3])
        True

    If *obj* is not iterable, return False::

        >>> is_iterable(1)
        False

    If *obj* is ``None``, return None:

        >>> assert is_iterable(None) is None

    By default, binary and text strings, dicts, generic aliases are not considered iterable:

        >>> is_iterable('foo')
        False

    If *base_type* is set, objects for which ``isinstance(obj, base_type)``
    returns ``True`` won't be considered iterable.

        >>> is_iterable([1, 2])  # list considered iterable by default
        True
        >>> is_iterable([1, 2], base_type=(list,))  # Treat list as a unit
        False

    Set *base_type* to ``None`` to avoid any special handling and treat objects
    Python considers iterable as iterable:

        >>> is_iterable('foo', base_type=None)
        True

    Based on always_iterable from more_itertools==11.0.1
    """
    if obj is None:
        return None

    if (base_type is not None) and isinstance(obj, base_type):
        return False

    try:
        iter(obj)
        return True
    except TypeError:
        return False



import collections.abc
import types
from typing import Any, Iterable, Union, List, get_args, get_origin

from more_itertools import always_iterable

from .get_container_type import get_container_type
from .is_mapping import is_mapping
from .is_collection import is_collection
from .is_tuple import is_tuple


def is_instance(
        val: Any,
        possible_types: type | types.GenericAlias | types.UnionType | Iterable[
                        type | types.GenericAlias | types.UnionType]
) -> bool:
    for type_ in always_iterable(possible_types, base_type=(type(List[int]), type(list[int]))):
        if type_ == Any:
            return True

        type_origin = get_origin(type_)
        type_args = get_args(type_)
        if not type_origin and isinstance(type_, type):  # basic check
            if isinstance(val, type_):
                return True
        elif type_origin is Union or isinstance(type_, types.UnionType):
            if is_instance(val, type_args):
                return True
        else:
            container_type = get_container_type(type_origin)
            if container_type is None or not isinstance(val, container_type):
                continue

            if issubclass(container_type, collections.abc.Mapping):
                if is_mapping(val, type_args):
                    return True

            elif container_type is tuple:  # tuple has specific typing behaviour
                if is_tuple(val, type_args):
                    return True

            elif issubclass(container_type, collections.abc.Collection):
                if is_collection(val, type_args):
                    return True
    return False



#  Copyright (C) 2024
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import re


def url_to_snake_case(url: str, replacer: str = '_') -> str:
    return re.sub(r'[^a-zA-Z0-9_]+', replacer, url).strip(replacer)



#  Copyright (C) 2024
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import itertools
import types
import typing


def get_non_generic_args(generic_alias: types.GenericAlias) -> tuple[type, ...]:
    return tuple(itertools.chain.from_iterable(
        get_non_generic_args(arg) if isinstance(arg, types.GenericAlias | types.UnionType) else (arg, )
        for arg in typing.get_args(generic_alias)
    ))
